	INCLUDE equates.h
	INCLUDE memory.h
	INCLUDE cart.h
	INCLUDE io.h
	INCLUDE 65c02s.h
	INCLUDE sound.h

	EXPORT lcd_init
	EXPORT LCD_reset_
	EXPORT _2000w
	EXPORT _2001w
	EXPORT _2002w
	EXPORT _2003w
	EXPORT vram_R
	EXPORT vram_W
	EXPORT debug_
	EXPORT AGBinput
	EXPORT NESinput
	EXPORT paletteinit
	EXPORT newframe
	EXPORT agb_pal
	EXPORT ppustate
	EXPORT _2026w
	EXPORT newX
	EXPORT gammavalue
	EXPORT palettebank
	EXPORT fpsenabled
	EXPORT FPSValue
	EXPORT vbldummy
	EXPORT vblankfptr
	EXPORT vblankinterrupt
	EXPORT ConvertScreen


 AREA rom_code, CODE, READONLY

;----------------------------------------------------------------------------
SVPalettes; RGB 24bit.

;Grey
	DCB 0xFF,0xFF,0xFF, 0xAA,0xAA,0xAA, 0x55,0x55,0x55, 0x00,0x00,0x00
;Yellow
	DCB 0xF3,0xFF,0x33, 0xC1,0xCE,0x22, 0x6F,0x7B,0x11, 0x00,0x00,0x00
;Green
	DCB 0xCC,0xFF,0xCC, 0x88,0xEE,0x88, 0x11,0xBB,0x11, 0x00,0x11,0x00
;Blue
	DCB 0xCC,0xCC,0xFF, 0x88,0x88,0xFF, 0x11,0x11,0xCC, 0x00,0x00,0x21
;----------------------------------------------------------------------------
lcd_init	;(called from main.c) only need to call once
;----------------------------------------------------------------------------
	mov addy,lr

	mov r1,#0xffffff00		;build chr decode tbl
	ldr r2,=CHR_DECODE
ppi0	mov r0,#0
	tst r1,#0x01
	orrne r0,r0,#0x0001
	tst r1,#0x02
	orrne r0,r0,#0x0002
	tst r1,#0x04
	orrne r0,r0,#0x0010
	tst r1,#0x08
	orrne r0,r0,#0x0020
	tst r1,#0x10
	orrne r0,r0,#0x0100
	tst r1,#0x20
	orrne r0,r0,#0x0200
	tst r1,#0x40
	orrne r0,r0,#0x1000
	tst r1,#0x80
	orrne r0,r0,#0x2000
	strh r0,[r2],#2
	adds r1,r1,#1
	bne ppi0


	mov r1,#REG_BASE
	mov r0,#0x0008
	strh r0,[r1,#REG_DISPSTAT]	;vblank en

	mov r0,#7
	strh r0,[r1,#REG_COLY]		;darkness setting for faded screens (bigger number=darker)
	mov r0,#9					;prio 1 chrbase 8000
	strh r0,[r1,#REG_BG0CNT]

	ldr r0,=0x3F3F
	strh r0,[r1,#REG_WININ]		;WinIN0/1, Everything enabled inside Windows
	sub r0,r0,#1				;r0=0x3f3e
	strh r0,[r1,#REG_WINOUT]	;WinOUT0/1, BG0 not enable outside Win0
	mov r0,#200					;endpos
	orr r0,r0,#40<<8			;startpos
	strh r0,[r1,#REG_WIN0H]		;WinO HPos
	mov r0,#160
	strh r0,[r1,#REG_WIN0V]		;WinO VPos

	add r0,r1,#REG_BG0HOFS		;DMA0 always goes here
	str r0,[r1,#REG_DM0DAD]
	mov r0,#1					;1 word transfer
	strh r0,[r1,#REG_DM0CNT_L]
	ldr r0,=DMA0BUFF+4			;dmasrc=
	str r0,[r1,#REG_DM0SAD]

	str r1,[r1,#REG_DM1DAD]		;DMA1 goes here
	mov r0,#1					;1 word transfer
	strh r0,[r1,#REG_DM1CNT_L]

	add r2,r1,#REG_IE
	mov r0,#-1
	strh r0,[r2,#2]		;stop pending interrupts
	ldr r0,=0x1091
	strh r0,[r2]		;key,vblank,timer1,serial interrupt enable
	mov r0,#1
	strh r0,[r2,#8]		;master irq enable

	ldr r1,=AGB_IRQVECT
	ldr r2,=irqhandler
	str r2,[r1]

	bx addy
;----------------------------------------------------------------------------
LCD_reset_	;called with CPU reset
;----------------------------------------------------------------------------
	mov r9,lr
	mov r0,#0
	strb r0,lcdctrl		;NMI, LCD & IRQ off
	strb r0,ppuctrl1	;screen off
	strb r0,ppustat		;flags off

	ldr r1,=DMA0BUFF	;clear SCROLLBUF+DMA0BUFF
	mov r2,#644
	bl filler_

	mov r0,#0x2440
	orr r0,r0,r0,lsl#16
	ldr r1,=DMA1BUFF	;clear DISPCNT+DMA1BUFF
	mov r2,#404/2
	bl filler_

	bl paletteinit

	mov lr,r9
;----------------------------------------------------------------------------
makeborder;		also called from UI.c
;----------------------------------------------------------------------------
	mov r1,#AGB_VRAM
	add r1,r1,#5*2
	mov r0,#0
	mov r2,#20
oloop
	mov r3,#24
bordloop
	strh r0,[r1],#2
	add r0,r0,#1
	subs r3,r3,#1
	bne bordloop
	add r0,r0,#8
	add r1,r1,#8*2
	subs r2,r2,#1
	bne oloop

	bx lr

;----------------------------------------------------------------------------
paletteinit;	r0-r3 modified.
;called by ui.c:  void paletteinit(void)
;----------------------------------------------------------------------------
	stmfd sp!,{r4-r7,lr}
	ldr r7,=SVPalettes
	ldr r1,palettebank	;Which color set (yellow, grey...)
	add r1,r1,r1,lsl#1	;r1 x 3
	add r7,r7,r1,lsl#2	;r7 + r1 x 4
	ldr r6,=agb_pal
	ldr r1,gammavalue	;gamma value = 0 -> 4
	mov r4,#4
nomap					;map rrrrrrrrggggggggbbbbbbbb  ->  0bbbbbgggggrrrrr
	ldrb r0,[r7],#1		;Red ready
	bl gammaconvert
	mov r5,r0

	ldrb r0,[r7],#1		;Green ready
	bl gammaconvert
	orr r5,r5,r0,lsl#5

	ldrb r0,[r7],#1		;Blue ready
	bl gammaconvert
	orr r5,r5,r0,lsl#10

	str r5,[r6],#4
	subs r4,r4,#1
	bpl nomap

	ldmfd sp!,{r4-r7,lr}
	bx lr

;----------------------------------------------------------------------------
gammaconvert;	takes value in r0(0-0xFF), gamma in r1(0-4),returns new value in r0=0x1F
;----------------------------------------------------------------------------
	rsb r2,r0,#0x100
	mul r3,r2,r2
	rsbs r2,r3,#0x10000
	rsb r3,r1,#4
	orr r0,r0,r0,lsl#8
	mul r2,r1,r2
	mla r0,r3,r0,r2
	mov r0,r0,lsr#13

	bx lr
;----------------------------------------------------------------------------
showfps_		;fps output, r0-r3=used.
;----------------------------------------------------------------------------
	ldrb r0,fpschk
	subs r0,r0,#1
	movmi r0,#59
	strb r0,fpschk
	bxpl lr					;End if not 60 frames has passed

	ldrb r0,fpsenabled
	tst r0,#1
	bxeq lr					;End if not enabled

	ldr r0,fpsvalue
	cmp r0,#0
	bxeq lr					;End if fps==0, to keep it from appearing in the menu
	mov r1,#0
	str r1,fpsvalue

	mov r1,#100
	swi 0x060000			;Division r0/r1, r0=result, r1=remainder.
	add r0,r0,#0x30
	strb r0,fpstext+5
	mov r0,r1
	mov r1,#10
	swi 0x060000			;Division r0/r1, r0=result, r1=remainder.
	add r0,r0,#0x30
	strb r0,fpstext+6
	add r1,r1,#0x30
	strb r1,fpstext+7
	

	adr r0,fpstext
	ldr r2,=DEBUGSCREEN
db1
	ldrb r1,[r0],#1
	orr r1,r1,#0x4100
	strh r1,[r2],#2
	tst r2,#15
	bne db1

	bx lr
;----------------------------------------------------------------------------
palettebank	DCD 0
gammavalue DCD 0
fpstext DCB "FPS:    "
fpsenabled DCB 0
fpschk	DCB 0
		DCB 0,0
;----------------------------------------------------------------------------
	AREA wram_code1, CODE, READWRITE
irqhandler	;r0-r3,r12 are safe to use
;----------------------------------------------------------------------------
	mov r2,#REG_BASE
	mov r3,#REG_BASE
	ldr r1,[r2,#REG_IE]!
	and r1,r1,r1,lsr#16	;r1=IE&IF
	ldrh r0,[r3,#-8]
	orr r0,r0,r1
	strh r0,[r3,#-8]

		;---these CAN'T be interrupted
		ands r0,r1,#0x80
		strneh r0,[r2,#2]		;IF clear
		bne serialinterrupt
		;---
		adr r12,irq0

		;---these CAN be interrupted
		ands r0,r1,#0x01
		ldrne r12,vblankfptr
		bne jmpintr
		ands r0,r1,#0x10
		ldrne r12,=timer1interrupt
		;----
		moveq r0,r1		;if unknown interrupt occured clear it.
jmpintr
	strh r0,[r2,#2]		;IF clear

	mrs r3,spsr
	stmfd sp!,{r3,lr}
	mrs r3,cpsr
	bic r3,r3,#0x9f
	orr r3,r3,#0x1f			;--> Enable IRQ & FIQ. Set CPU mode to System.
	msr cpsr_cf,r3
	stmfd sp!,{lr}
	adr lr,irq0

	mov pc,r12


irq0
	ldmfd sp!,{lr}
	mrs r3,cpsr
	bic r3,r3,#0x9f
	orr r3,r3,#0x92        		;--> Disable IRQ. Enable FIQ. Set CPU mode to IRQ
	msr cpsr_cf,r3
	ldmfd sp!,{r0,lr}
	msr spsr_cf,r0
vbldummy
	bx lr
;----------------------------------------------------------------------------
vblankfptr DCD vbldummy			;later switched to vblankinterrupt
twitch DCD 0
vblankinterrupt;
;----------------------------------------------------------------------------
	stmfd sp!,{r4-r7,globalptr,lr}
	ldr globalptr,=|wram_globals0$$Base|

	bl showfps_


	ldr r2,=DMA0BUFF	;setup DMA buffer for scrolling:
	add r3,r2,#160*4
	ldr r1,dmascrollbuff

vbl6
	ldmia r1!,{r0,r4-r7}
	stmia r2!,{r0,r4-r7}
	cmp r2,r3
	bmi vbl6

	ldr r3,=DISPCNTBUFF
	ldr r4,=BG0CNTBUFF

	mov r1,#REG_BASE
	strh r1,[r1,#REG_DM0CNT_H]		;DMA stop
	strh r1,[r1,#REG_DM1CNT_H]
	strh r1,[r1,#REG_DM3CNT_H]

	ldr r0,=DMA0BUFF				;setup HBLANK DMA for display scroll:
	ldr r2,[r0],#4
	str r2,[r1,#REG_BG0HOFS]		;set 1st value manually, HBL is AFTER 1st line
	ldr r0,=0xA660					;noIRQ hblank 32bit repeat incsrc inc_reloaddst
	strh r0,[r1,#REG_DM0CNT_H]		;DMA go
									;setup HBLANK DMA for DISPCNT (BG/OBJ enable)
	ldrh r2,[r3],#2
	strh r2,[r1,#REG_DISPCNT]		;set 1st value manually, HBL is AFTER 1st line
	str r3,[r1,#REG_DM1SAD]			;dmasrc=
	ldr r0,=0xA240					;noIRQ hblank 16bit repeat incsrc fixeddst
	strh r0,[r1,#REG_DM1CNT_H]		;DMA go
									;setup HBLANK DMA for BG CHR
;	add r0,r1,#REG_BG0CNT
;	str r0,[r1,#REG_DM3DAD]
;	ldr r2,[r4],#2
;	strh r2,[r1,#REG_BG0CNT]
;	str r4,[r1,#REG_DM3SAD]
;	ldr r0,=0xA2400001				;noIRQ hblank 16bit repeat incsrc fixeddst, 1 word transfer
;	str r0,[r1,#REG_DM3CNT_L]		;DMA go

	ldmfd sp!,{r4-r7,globalptr,pc}

;----------------------------------------------------------------------------
newframe	;called at line 0	(r0-r9 safe to use)
;----------------------------------------------------------------------------
	str lr,[sp,#-4]!

;-----------------------
	ldr r0,ctrl1old
	ldr r1,ctrl1line
	mov addy,#159
	bl ctrl1finish
	mov r0,#0
	str r0,ctrl1line
;-----------------------
	ldr r0,scrollXold
	ldr r1,scrollXline
	mov addy,#159
	bl scrollXfinish
	mov r0,#0
	str r0,scrollXline
;--------------------------
	ldr r0,scrollbuff
	ldr r1,dmascrollbuff
	str r1,scrollbuff
	str r0,dmascrollbuff

	bl ConvertScreen

	mov r8,#AGB_PALETTE		;palette transfer
	adrl addy,agb_pal
nf8	ldmia addy!,{r0-r7}
	stmia r8,{r0,r1}
	add r8,r8,#32
	stmia r8,{r2,r3}
	add r8,r8,#32
	stmia r8,{r4,r5}
	add r8,r8,#32
	stmia r8,{r6,r7}
	add r8,r8,#0x1a0
	tst r8,#0x200
	bne nf8			;(2nd pass: sprite pal)

	ldr pc,[sp],#4
;----------------------------------------------------------------------------
_2000w
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
_2001w
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
_2002w
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
_2003w
;----------------------------------------------------------------------------
	mov pc,lr

;----------------------------------------------------------------------------
_2026w		;LCD & IRQs
;----------------------------------------------------------------------------
	ldrb r1,lcdctrl
	strb r0,lcdctrl
	eor r1,r0,r1
	tst r1,#0xE0
	movne r0,r0,lsr#5
	movne addy,lr
	blne map89AB_
	movne lr,addy
	ldrb r0,lcdctrl

	mov r1,#0x2440		;WIN0, BG2 enable. DISPCNTBUFF startvalue. 0x2440
	tst r0,#0x08		;lcd en?
	orrne r1,r1,#0x0100

	adr r2,ctrl1old
	swp r0,r1,[r2]		;r0=lastval

	adr r2,ctrl1line
	ldr addy,scanline	;addy=scanline
	cmp addy,#159
	movhi addy,#159
	swp r1,addy,[r2]	;r1=lastline, lastline=scanline
ctrl1finish
	ldr r2,=DISPCNTBUFF
	add r1,r2,r1,lsl#1
	add r2,r2,addy,lsl#1
ct1	strh r0,[r2],#-2	;fill backwards from scanline to lastline
	cmp r2,r1
	bpl ct1

	mov pc,lr

ctrl1old	DCD 0x2440	;last write
ctrl1line	DCD 0 ;when?
;----------------------------------------------------------------------------
;bgscroll_W	;(2005)
;----------------------------------------------------------------------------
;bgscrollX
	strb r0,scrollX
newX			;ctrl0_W, loadstate jumps here
	ldr r0,scrollX
	adr r1,scrollXold
	swp r0,r0,[r1]		;r0=lastval

	adr r2,scrollXline
	ldr addy,scanline	;addy=scanline
	cmp addy,#159
	movhi addy,#159
	swp r1,addy,[r2]	;r1=lastline, lastline=scanline
scrollXfinish		;newframe jumps here
	ldr r2,scrollbuff
	add r1,r2,r1,lsl#2
	add r2,r2,addy,lsl#2
sx1	strh r0,[r2],#-4	;fill backwards from scanline to lastline
	cmp r2,r1
	bpl sx1
	mov pc,lr

scrollXold DCD 0 ;last write
scrollXline DCD 0 ;..was when?
;----------------------------------------------------------------------------

bgscrollY
	strb r0,scrollY

	mov pc,lr
;----------------------------------------------------------------------------
vram_R	;(4000...)
;----------------------------------------------------------------------------
	bic addy,addy,#0x1E000
	ldr r2,=NES_VRAM
	ldrsb nes_nz,[r2,addy]
	mov pc,lr
;----------------------------------------------------------------------------
vram_W	;(4000...)
;----------------------------------------------------------------------------
	bic addy,addy,#0xFE000
	ldr r2,=NES_VRAM
	strb r0,[r2,addy]
	mov pc,lr
;----------------------------------------------------------------------------
ConvertScreen	;
;----------------------------------------------------------------------------
	stmfd sp!,{r3-r7}

	adr r4,chr_decode
	ldr r2,=NES_VRAM
	mov r3,#AGB_VRAM		;AGB BG tileset
	add r3,r3,#0x8000		;AGB BG tileset2



	mov r7,#20				;20 tiles high screen
ScLoop
	mov r6,#8				;8 pix high tiles
TiLoop
	mov r5,#24				;192 pix
RwLoop
	ldrb r0,[r2],#1	;read 4 1st pixels
	mov r0,r0,lsl#1
	ldrh r0,[r4,r0]
	ldrb r1,[r2],#1	;read 4 more pixels
	mov r1,r1,lsl#1
	ldrh r1,[r4,r1]
	orr r0,r0,r1,lsl#16
	str r0,[r3],#32
	subs r5,r5,#1
	bne RwLoop

	sub r3,r3,#32*24
	add r3,r3,#4
	subs r6,r6,#1
	bne TiLoop

	add r3,r3,#32*31
	subs r7,r7,#1
	bne ScLoop



	ldmfd sp!,{r3-r7}
	mov pc,lr

;----------------------------------------------------------------------------
debug_		;debug output, r0=val, r1=line, r2=used.
;----------------------------------------------------------------------------
 [ DEBUG
	ldr r2,=DEBUGSCREEN
	add r2,r2,r1,lsl#6
db0
	mov r0,r0,ror#28
	and r1,r0,#0x0f
	cmp r1,#9
	addhi r1,r1,#7
	add r1,r1,#0x30
	orr r1,r1,#0x4100
	strh r1,[r2],#2
	tst r2,#15
	bne db0
 ]
	bx lr

;----------------------------------------------------------------------------

agb_pal		% 32*2	;copy this to real AGB palette every frame

scrollbuff DCD SCROLLBUFF1
dmascrollbuff DCD SCROLLBUFF2

;----------------------------------------------------------------------------
	AREA wram_globals1, CODE, READWRITE
FPSValue
	DCD 0
AGBinput		;this label here for main.c to use
	DCD 0 ;AGBjoypad (why is this in ppu.s again?  um.. i forget)
NESinput	DCD 0 ;NESjoypad (this is what NES sees)
	DCD 2 ;adjustblend
ppustate
	DCD 0 ;vramaddr
	DCD 0 ;vramaddr2 (temp)
	DCD 0 ;scrollX
	DCD 0 ;scrollY
	DCD 0 ;timerlatch
	DCD 0 ;timerval

	DCB 0 ;sv_ioport
	DCB 0 ;irq_r
	DCB 0 ;ppustat
	DCB 0 ;toggle
	DCB 0 ;lcdctrl
	DCB 0 ;ppuctrl1
;...update load/savestate if you move things around in here
;----------------------------------------------------------------------------
	END
